import asyncio
import json
from domsync import Document

class DomsyncServer():
    """
    :class:`domsync.domsync_server.DomsyncServer` is a websocket server. When a client connects, it creates a :class:`domsync.Document` instance for the client
    and calls the ``connection_handler`` function that was provided on initialisation (see below). The client's document can be accessed with :meth:`get_document`.
    After the document has been manipulated and :meth:`flush` is called, the Javascript code that got generated by the document as a result of the changes is sent to
    the client where it gets applied to the Browser DOM and the changes appear on the screen.
    Any callback methods that were registered on the client's document with :meth:`domsync.core._Element.addEventListener` are also handled by the server: whenever
    those methods trigger ``ws_send`` on the client side, the message containig the event is received by the server and the corresponding Python callback function is triggered.

    :param connection_handler: is a callback function that is called each time a client connects to the server.
       It has two arguments: the first contains the DomsyncServer instance, the second contains the client websocket connection instance.
    :type connection_handler: Callable(:class:`domsync.domsync_server.DomsyncServer`, ``WebSocketServerProtocol``)

    :param host: host name to listen on
    :type host: str

    :param port: port number to listen on
    :type port: int

    :param verbose: optional, should we print status messages to stdout? default = True.
    :type verbose: bool

    :param root_id: optional, id of the element in the client-side HTML where domsync should be rendered. default = 'domsync_root_id'.
    :type root_id: str
    """
    def __init__(self, connection_handler, host, port, verbose = True, root_id = 'domsync_root_id'):
        self.host = host
        self.port = port
        self.verbose = verbose
        self.root_id = root_id
        self.clients = {} # client websocket instance -> domsync Document
        self.connection_handler = connection_handler

    async def serve(self):
        """
        starts the server

        :returns: None
        """
        import websockets
        self.server = await websockets.serve(self._on_ws_client_connect, self.host, self.port)
        if self.verbose: print(f'domsync server started on ws://{self.host}:{self.port}')

    async def close(self):
        """
        stops the server

        :returns: None
        """
        assert self.server.is_serving()
        self.server.close()
        await self.server.wait_closed()
        assert not self.server.is_serving()

    async def _on_ws_client_connect(self, client, _path):
        import websockets
        assert client not in self.clients
        doc = Document(self.root_id)
        self.clients[client] = doc

        asyncio.create_task(self.connection_handler(self, client))

        while True:
            try:
                msg = await client.recv()
            except websockets.exceptions.ConnectionClosedOK:
                break

            try:
                msg = json.loads(msg)
            except json.decoder.JSONDecodeError:
                continue

            if msg.get('domsync'):
                doc.handle_event(msg)
                await self.flush(client)

        del self.clients[client]
    
    def is_connected(self, client):
        """
        returns whether the given client is still connected

        :param client: a websocket client connection instance
        :type: client: ``WebSocketServerProtocol``

        :returns: True if the client is still connected, False otherwise
        :rtype: bool
        """
        return client in self.clients

    def get_clients(self):
        """
        :returns: list of connected clients
        :rtype: list of ``WebSocketServerProtocol``
        """
        return list(self.clients.keys())

    def get_document(self, client):
        """
        returns the document of the given client

        it is recommended to check :meth:`domsync.domsync_server.DomsyncServer.is_connected` immediately before calling this function because this function throws an exception if the client has disconnected

        :param client: a websocket client connection instance
        :type: client: ``WebSocketServerProtocol``

        :returns: the document associated with the client connection
        :rtype: :class:`domsync.Document`
        """
        return self.clients[client]

    async def flush(self, client):
        """
        sends the generated Javascript code updates to the client that happened due to manipulations to the client's document since the last time this function was called.

        it is recommended to check :meth:`domsync.domsync_server.DomsyncServer.is_connected` immediately before calling this function because this function throws an exception if the client has disconnected

        :param client: the client to send the updates for
        :type: client: ``WebSocketServerProtocol``

        :returns: None
        """
        doc = self.clients[client]
        js = doc.render_js_updates()
        if len(js) > 0:
            await client.send(js)

    # async def flush_all(self):
    #     """
    #     similar to :meth:`domsync.domsync_server.DomsyncServer.flush`, but sends updates to all connected clients

    #     :returns: None
    #     """
    #     for client in self.clients:
    #         doc = self.clients[client]
    #         js = doc.render_js_updates()
    #         if len(js) > 0:
    #             await client.send(js)



# copied from https://way2tutorial.com/html/tag/index.php
_valid_tags = ["a","abbr","address","area","b","base","bdo","blockquote","body","br","button","caption","cite","code","col","colgroup","dd","del","dfn","div","dl","dt","em","fieldset","form","h1","h2","h3","h4","h5","h6","head","hr","html","i","iframe","img","input","ins","kbd","label","legend","li","link","map","menu","meta","noscript","object","ol","optgroup","option","p","param","pre","q","s","samp","script","select","small","span","strong","style","sub","sup","table","tbody","td","textarea","tfoot","th","thead","title","tr","u","ul","var"]

# https://www.w3schools.com/jsref/dom_obj_event.asp
_valid_events = ["abort","afterprint","animationend","animationiteration","animationstart","beforeprint","beforeunload","blur","canplay","canplaythrough","change","click","contextmenu","copy","cut","dblclick","drag","dragend","dragenter","dragleave","dragover","dragstart","drop","durationchange","ended","error","focus","focusin","focusout","fullscreenchange","fullscreenerror","hashchange","input","invalid","keydown","keypress","keyup","load","loadeddata","loadedmetadata","loadstart","message","mousedown","mouseenter","mouseleave","mousemove","mouseover","mouseout","mouseup","mousewheel","offline","online","open","pagehide","pageshow","paste","pause","play","playing","popstate","progress","ratechange","resize","reset","scroll","search","seeked","seeking","select","show","stalled","storage","submit","suspend","timeupdate","toggle","touchcancel","touchend","touchmove","touchstart","transitionend","unload","volumechange","waiting","wheel",]

class _Element(dict): # _Element is private because we are only meant to create an instance through Document.createElement
    """:class:`domsync.core._Element` is analogous to the Javascript Element which represents an individual HTML element.
    The name of the class starts with an underscore, expressing the fact that this class should not be instantiated by the user,
    instead all instances of this class are created by :meth:`domsync.Document.createElement`.

    :param document: document to create the element within
    :type document: :class:`domsync.Document`

    :param id: unique id of the element
    :type id: str

    :param tagName: tag name of the element
    :type tagName: str

    :class attributes:

    * **innerText** - sets and gets the element's innerText, analogous to Javascript element.innerText
    * **value** - sets and gets the element's value, analogous to Javascript element.value
    * **tagName** - gets the element's tagName, analogous to Javascript element.tagName
    * **children** - gets the element's list of child elements, analogous to Javascript element.children
    * **firstElementChild** - gets the element's first child element, analogous to Javascript element.firstElementChild
    * **lastElementChild** - gets the element's last child element, analogous to Javascript element.lastElementChild
    * **attributes** - gets the element's dictionary of attributes
    * **id** - gets the element's id

    """
    def __init__(self, document, id, tagName):
        assert tagName in _valid_tags
        super(_Element, self).__init__({
            'document': document,
            'id': id,
            'tag': tagName,
            'children': [],
            'parent': None,
            'attributes': {},
            'innerText': None,
            'value': None,
        })

    def _js_push(self, line):
        self['document']._js_push(line)

    def appendChild(self, el_child):
        """
        analogous to Javascript Element.appendChild

        :param el_child: child element to append
        :type el_child: :class:`domsync.core._Element`

        :returns: None
        """
        assert isinstance(el_child, _Element)
        assert not self.get('innerHTML_flag',False), "this element has some innerHTML, remove it first by self.innerHTML = "" before adding children"
        assert el_child is self['document'].getElementById(el_child['id'])
        assert el_child.parentElement is None, "child is already under a parent"
        el_child['parent'] = self
        self['children'].append(el_child)
        self._js_push(f"""__domsync__["{self['id']}"].appendChild(__domsync__["{el_child['id']}"]);\n""")

    def insertBefore(self, el_child_to_insert, el_child_before):
        """
        inserts an element as a child before an existing child element

        analogous to Javascript Element.insertBefore

        :returns: None
        """
        assert isinstance(el_child_to_insert, _Element)
        assert not self.get('innerHTML_flag',False), "this element has some innerHTML, remove it first by self.innerHTML = "" before adding children"
        assert el_child_to_insert.parentElement is None, "child is already under a parent"
        el_child_to_insert['parent'] = self
        found = False
        for i in range(len(self['children'])):
            if self['children'][i] is el_child_before:
                found = True
                break
        assert found
        self['children'].insert(i, el_child_to_insert)
        self._js_push(f"""__domsync__["{self['id']}"].insertBefore(__domsync__["{el_child_to_insert['id']}"], __domsync__["{el_child_before['id']}"]);\n""")

    def remove(self):
        """
        removes the element

        analogous to Javascript Element.remove

        :returns: None
        """
        _id = self['id']
        assert _id in self['document']['elements_by_id']
        parent = self.parentElement
        len_before = len(parent['children'])
        parent['children'] = [el for el in parent['children'] if el['id']!=_id]
        len_after = len(parent['children'])
        assert len_after == len_before - 1, len_before
        self._js_push(f"""__domsync__["{self['id']}"].remove();\n""")

        del self['document']['elements_by_id'][_id]
        if _id in self['document']['callbacks']:
            del self['document']['callbacks'][_id]
        to_remove = [el['id'] for el in self['children']]
        while len(to_remove):
            _id = to_remove.pop(0)
            element = self['document']['elements_by_id'][_id]
            del self['document']['elements_by_id'][_id]
            if _id in self['document']['callbacks']:
                del self['document']['callbacks'][_id]
            to_remove.extend([el['id'] for el in element['children']])

    def getAttribute(self, attrib, default = None):
        """
        gets an attribute of an element

        analogous to Javascript Element.getAttribute

        :param attrib: name of the attribute to return
        :type attrib: str

        :param default: Optional, value to return in case the attribute is not available

        :returns: str, the value of the attribute if the attribute exists, otherwise the value of the ``default`` argument
        """
        assert attrib != 'id' and type(attrib) is str
        return self['attributes'].get(attrib, default)

    def setAttribute(self, attrib, value):
        """
        sets the attribute of an element

        analogous to Javascript Element.setAttribute

        :param attrib: name of the attribute to set
        :type attrib: str

        :param value: value of the attribute to set
        :type value: str

        :returns: None
        """
        assert str_is_safe(attrib)
        assert attrib != 'id' and type(attrib) is str and type(value) is str
        if attrib.startswith('on'): assert attrib[2:] not in _valid_events, "please use addEventListener to add an event"
        if attrib.startswith('on'):
            # https://stackoverflow.com/questions/97578/how-do-i-escape-a-string-inside-javascript-code-inside-an-onclick-handler
            # need to escape quotes in code string
            value = str_escape_for_js(value)
        if self['attributes'].get(attrib) != value:
            self['attributes'][attrib] = value
            self._js_push(f"""__domsync__["{self['id']}"].setAttribute("{attrib}","{value}");\n""")

    def removeAttribute(self, attrib):
        """
        removes the given attribute

        analogous to Javascript Element.removeAttribute

        :param attrib: name of the attribute to remove
        :type attrib: str

        :returns: None
        """
        assert attrib != 'id' and type(attrib) is str
        del self['attributes'][attrib]
        self._js_push(f"""__domsync__["{self['id']}"].removeAttribute("{attrib}");\n""")

    def addEventListener(self, event, callback, js_value_getter = None):
        """
        adds an event listener to the element

        analogous to Javascript addEventListener

        :param event: name of the event to listen to, see https://www.w3schools.com/jsref/dom_obj_event.asp for a list of valid events
        :type event: str

        :param callback: | the callback function to be called when the event happens. the function must take one argument which is a dict containing the details of the message:
                         | 'event': name of the event, one of https://www.w3schools.com/jsref/dom_obj_event.asp
                         | 'id': id of the element that the event happened on
                         | 'doc': :class:`domsync.Document` instance
                         | 'value': value returned as a result of evaluating js_value_getter (see below)
        :type callback: Callable(dict)

        :param js_value_getter: | a javascript expression that is executed in the context of the event and the return value of which is retrned in the 'value' field of the event message.
                                | A good example is ``'this.value'`` in case of the 'input' event of an <input type='text'> element, this will retun the current changed value of the input.
                                | Another example would be ``'this.options[this.selectedIndex].value'`` in case of the 'input' event of a <select> element, this will retun the current changed value of the selection.
                                | In case of a 'click' event of a <button> there is no need to specify a js_value_getter because the click event doesn't carry any relevant value (apart form the fact that the event happened).
        :type js_value_getter: str

        :returns: None
        """
        assert event in _valid_events
        event_msg = {
            'domsync':True,
            'event':event,
            'id': self['id'],
            'value': js_value_getter,
        }
        import json
        event_msg = json.dumps(event_msg)
        if js_value_getter is not None:
            event_msg = event_msg.replace('"'+js_value_getter+'"',js_value_getter)
        callback_js = r"function(){ws_send("+event_msg+r")}"
        self['document']._register_callback(self['id'], event, callback)
        self._js_push(f"""__domsync__["{self['id']}"].addEventListener("{event}",{callback_js});\n""")

    def _setInnerHTML(self, innerHTML):
        """
        analogous to Javascript Element.innerHTML = innerHTML
        """
        assert type(innerHTML) is str
        # NOTE: in our representation we store innerHTML in self['innerText']
        #       also, we allow no innerHTML setting if the element has children
        assert len(self['children']) == 0, "cannot have children if setting innerHTML"
        if innerHTML != self['innerText']:
            self['innerText'] = innerHTML
            self['innerHTML_flag'] = True if len(innerHTML)>0 else False
            self._js_push(f"""__domsync__["{self['id']}"].innerHTML = "{innerHTML}";\n""")

    def _getInnerHTML(self):
        """
        analogous to Javascript Element.innerHTML
        """
        assert self['innerHTML_flag'] == True, "can only access innerHTML if it was set with innerHTML"
        return self['innerText']

    def _setText(self, text):
        """
        analogous to Javascript Element.innerText = text
        """
        assert type(text) is str, "we don't allow any other types than str to be stored in the DOM because we didn't want to make parsing/rendering/formatting part of the DOM, that should happen outside"
        if text != self['innerText']:
            self['innerText'] = text
            self._js_push(f"""__domsync__["{self['id']}"].innerText = `{text}`;\n""")

    def _setValue(self, value):
        """
        analogous to Javascript Element.value = value
        """
        assert type(value) is str, "we don't allow any other types than str to be stored in the DOM because we didn't want to make parsing/rendering/formatting part of the DOM, that should happen outside"
        if value != self['value']:
            self['value'] = value
            self._js_push(f"""__domsync__["{self['id']}"].value = `{value}`;\n""")

    def __setattr__(self, name, value):
        if name == 'innerHTML':
            self._setInnerHTML(value)
        elif name == 'innerText':
            self._setText(value)
        elif name == 'value':
            self._setValue(value)
        else:
            raise Exception('unsupported attribute: ' + str(name) + ' of type ' + str(type(name)))
    
    def __getattr__(self, name):
        if name == 'innerHTML':
            return self._getInnerHTML()
        elif name == 'innerText':
            return self['innerText']
        elif name == 'value':
            return self['value']
        elif name == 'tagName':
            return self['tag']
        elif name == 'children':
            return self['children']
        elif name == 'firstElementChild':
            return self['children'][0]
        elif name == 'lastElementChild':
            return self['children'][-1]
        elif name == 'attributes':
            return self['attributes']
        elif name == 'id': # NOTE: this is not in JS, only for us for convenience. in JS should be self.getAttribute('id')
            return self['id']
        elif name == 'parentElement':
            return self['parent']
        else:
            raise Exception('unsupported attribute: ' + str(name) + ' of type ' + str(type(name)))

class Document(dict):
    """:class:`domsync.Document` is analogous to the Javascriot DOM document which contains a tree of :class:`domsync.core._Element` objects.
    Every manipulation to the document generates Javascript code that when sent to the Browser client and evaluated results in the same DOM changes on the client side
    that happened in the :class:`domsync.Document`.

    :param root_id: id of the element in the client-side HTML where domsync should render
    :type root_id: str
    """
    def __init__(self, root_id):
        """Constructor method
        """
        root_tag = 'div' # doesn't matter what the tag of the root element actually is, we store it as div in our representatiopn just to have a valid tag
        root_el = _Element(self, root_id, root_tag)
        super(Document, self).__init__({
            'elements_by_id': { # returns the element of an id
                root_id: root_el,
            },
            'js_buffer': [],
            'id_autoinc': 0,
            'root_id': root_id,
            'callbacks': {},
        })
        self._js_push("var __domsync__ = [];\n")
        self._js_push(f"""__domsync__["{root_id}"] = document.getElementById("{root_id}");\n""")

    def _js_push(self, line):
        self['js_buffer'].append(line)

    def _get_autoinc_id(self):
        _id = '__domsync_el_'+str(self['id_autoinc'])
        self['id_autoinc'] += 1
        return _id

    def getRootId(self):
        """Returns the id of the root element that was passed on initialisation.

        :return: the id of the root element
        :rtype: str
        """
        return self['root_id']

    def getElementById(self, id, strict=True):
        """
        Returns the element of the given ID

        analogous to Javascript document.getElementById

        :param id: id of the element
        :type id: str

        :param strict: optional, if True (default), then an exception is thrown if the provided id doesn't exist. if False, then None is returned if the id doesn't exist
        :type strict: bool

        :return: the element of the provided id or None if the element doesn't exist and strict is False
        :rtype: :class:`domsync.core._Element`
        """
        if strict:
            assert id in self['elements_by_id'], "unknown id: " + str(id)
            return self['elements_by_id'][id]
        else:
            return self['elements_by_id'].get(id)

    def getElementsByClassName(self, className):
        """
        Returns a list of elements with the specified class name

        analogous to Javascript document.getElementsByClassName

        :param className: class name of the element
        :type className: str

        :return: the elements of the given class name
        :rtype: list of :class:`domsync.core._Element`
        """
        classNames = className.split(' ')
        return [el for el in self['elements_by_id'].values() if el.getAttribute('class') in classNames]

    def getElementsByTagName(self, tagName):
        """
        Returns a list of elements with the specified tag name

        analogous to Javascript document.getElementsByTagName

        :param tagName: tag name of the element
        :type tagName: str

        :return: the elements of the given tag
        :rtype: list of :class:`domsync.core._Element`
        """
        return [el for el in self['elements_by_id'].values() if (el.tagName == tagName or tagName == '')]

    def createElement(self, tagName, id=None, innerText=None, value=None, attributes=None):
        """
        Creates a new element in the :class:`domsync.Document` but doesn't add it as a child to any existing elements, that needs to be done separately using ``appendChild``.

        analogous to Javascript document.createElement

        This is the only way to create a new element because each element needs to be registered with the :class:`domsync.Document`.
        This is the reason why the name of :class:`domsync.core._Element` starts with an  underscore character signalling that it's a private class that is not meant
        to be instantiated by the user.

        :param tagName: tag name of the element to be created
        :type tagName: str

        :param id: Optional, unique id of the new element. if not provided, an automatically generated unique id will be used.
                   Note that this is a difference between domsync and Javascript: in domsync every element has a unique id while in Javascript not necessarily.
        :type id: str

        :param innerText: Optional, if provided this will be set as the innerText of the newly created element.
                          equivalent to later setting the .innerText attribute of the element.
        :type innerText: str

        :param value: Optional, if provided this will be set as the value attribute of the newly created element.
                      Typically only used in case of <input type='text'> elements.
                      equivalent to later setting the .value attribute of the element.
        :type value: str

        :param attributes: Optional, if provided these attributes will be set on the element.
                        equivalent to later using the setAttribute method on the element.
        :type attributes: dict

        :return: the newly created element
        :rtype: :class:`domsync.core._Element`
        """
        if id is None:
            id = self._get_autoinc_id()
        assert id not in self['elements_by_id']
        el = _Element(self, id, tagName)
        self['elements_by_id'][id] = el
        self._js_push(f"""__domsync__["{el['id']}"]=document.createElement("{el['tag']}");__domsync__["{el['id']}"].setAttribute("id","{el['id']}");\n""")
        if innerText is not None:
            assert type(innerText) is str
            el.innerText = innerText
        if value is not None:
            el.value = value
        if attributes is not None:
            assert type(attributes) is dict
            for attrib, value in attributes.items():
                el.setAttribute(attrib, value)
        return el

    def render_js_updates(self):
        """
        Returns the Javascript code changes that accummulated in the internal buffer of the :class:`domsync.Document` due to any manipulations since the last call to this function.
        This is the method for generating the Javascript code updates that can be sent to the client.
        :class:`domsync.domsync_server.DomsyncServer` uses this behind the scenes, so you only need to deal with this function if you want to use your own server.

        :return: the Javascript code generated since the last call to this function.
        :rtype: str
        """
        res = self['js_buffer']
        self['js_buffer'] = []
        return ''.join(res)

    def render_js_full(self):
        """
        Returns a full snapshot of Javascript code that represents the current state of the :class:`domsync.Document` from scratch, not just the updates since the last time.
        It is useful when you have one :class:`domsync.Document` instance in your memory and want to show the same instance to every users, like for example a read-only dashboard.
        In that case you can update your :class:`domsync.Document` and send updates to already connected clients using the ``render_js_updates`` method, but whenever a new client connects
        you can use this method to send an initial full snapshot of the current state of the doc. Again, this only needs to be used if you decide to keep one :class:`domsync.Document`
        instance for all users, as opposed to one :class:`domsync.Document` for each user.
        If you use :class:`domsync.domsync_server.DomsyncServer`, you don't need to deal with this method because the server maintains one :class:`domsync.Document`
        for each client and send updates behind the scenes automatically anyways.
        You only need to deal with this method if you decide to use your own server.

        :return: Javascript code containnig the full current state of the document.
        :rtype: str
        """
        assert self['js_buffer'] == [], 'can only call render_js_full right after render_js_updates'
        new_doc = Document(self.getRootId())
        ids_to_copy = [el.id for el in self.getElementById(self.getRootId()).children]
        while len(ids_to_copy):
            id = ids_to_copy.pop(0)
            old_el = self.getElementById(id)
            ids_to_copy.extend([el.id for el in old_el.children])
            parent_id = old_el.parentElement.id
            assert type(parent_id) is str
            new_doc.createElement(old_el.tagName, id=old_el.id, innerText=old_el.innerText, value=old_el.value, attributes=old_el.attributes)
            new_el = new_doc.getElementById(id)
            for event, callback in self['callbacks'].get(id,{}).items():
                new_el.addEventListener(event, callback)
            new_parent = new_doc.getElementById(parent_id)
            new_parent.appendChild(new_el)
        return new_doc.render_js_updates()

    def handle_event(self, msg):
        """
        The way event listeners work with domsync is that we set the event listener of an element using :meth:`domsync.core._Element.addEventListener` which
        on the client side causes the ``ws_send`` Javascript function to be executed which sends back a message to the server containing the
        details of the event. When that message arrives to the server, this method needs to be called with the message as an argument which
        eventually triggers the callback function to be executed that was added using the :meth:`domsync.core._Element.addEventListener` method.

        You only need to deal with this if you are using your own server. If you use :class:`domsync.domsync_server.DomsyncServer`,
        this is called automatically behind the scenes.

        :param msg: | is an event message generated on the client side containing:
                    | 'id' of the element that generated the event
                    | 'event' name of the HTML event that happened (see https://www.w3schools.com/jsref/dom_obj_event.asp for the list of events)
                    | 'value' associated with the event as defined when the event was added using the ``js_value_getter`` argument of :meth:`domsync.core._Element.addEventListener`
        :type msg: dict

        :returns: whatever the callback function returns that was added using :meth:`domsync.core._Element.addEventListener`
        """
        assert msg['domsync']
        if msg['event'] in self['callbacks'].get(msg['id'],{}):
            msg['doc'] = self
            callback = self['callbacks'][msg['id']][msg['event']]
            return callback(msg)

    def _register_callback(self, id, event, callback):
        """
        use this function to register an event handler callback
        """
        assert id in self['elements_by_id']
        self['callbacks'].setdefault(id,{})
        assert event not in self['callbacks'][id]
        self['callbacks'][id][event] = callback

def str_is_safe(s):
    return '"' not in s and "'" not in s and "`" not in s

def str_escape_for_js(s):
    return s.replace("'",r"\x27").replace('"',r"\x22")


from setuptools import setup

import pathlib
import re

root_dir = pathlib.Path(__file__).parent

long_description = (root_dir / 'README.md').read_text(encoding='utf-8')

# PyPI disables the "raw" directive.
long_description = re.sub(
    r"^\.\. raw:: html.*?^(?=\w)",
    "",
    long_description,
    flags=re.DOTALL | re.MULTILINE,
)

setup(
    name='domsync',
    version='0.1.0.0.1',
    url='https://github.com/pdivos/domsync',
    description='Python server-side DOM synchronised with Browser',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='pdivos',
    author_email='pdivos@gmail.com',
    license='BSD',
    packages=['domsync'],
    package_dir={'':'src'},
    install_requires=[
        'websockets',
    ],
    test_suite="tests",
    zip_safe=False,
    project_urls = {
        'Documentation': 'https://domsync.readthedocs.io/',
        # Changelog = https://websockets.readthedocs.io/en/stable/project/changelog.html
        # Tracker = https://github.com/aaugustin/websockets/issues
    },

)


from domsync.core import Document

class Component(dict):
    """
    a Component is a reusable group of Elements along with some private data
    each Component has exactly one root Element which may have child Elements
    upon __init__ the Component writes it's Elements in the Document, under an existing parent Element
    """
    def __init__(self, doc, parent_id, id = None):
        """
        :param doc: Document to add the components' elements into
        :param parent_id: id of an Element in the Document under which the COmponent's Elements shall be added
        :param id: id of the root Element inserted by the Component
        """
        assert type(doc) is Document and type(parent_id) is str and parent_id in doc['elements_by_id']
        if id is None:
            id = doc._get_autoinc_id()
        super(Component, self).__init__({
            'doc': doc,
            'parent_id': parent_id,
            'id': id,
        })

    def getDoc(self):
        return self['doc']

    def getElement(self):
        """
        :returns: the root Element of the Component
        """
        return self['doc'].getElementById(self['id'])

    def getRootId(self):
        """
        :returns: the id of the root Element of the Component
        """
        return self['id']

class TableComponent(Component):
    def __init__(self, doc, parent_id, columns, id=None):
        """
        :param doc: Document to add the components' elements into
        :param parent_id: id of an Element in the Document under which the COmponent's Elements shall be added
        :param columns: dict of column id -> column name. insert order in dict determines the order of columns.
                        or list of column names which are also the column ids.
        :param id: id of the root Element inserted by the Component
        """
        if type(columns) is list:
            columns = {el:el for el in columns}
        super(TableComponent, self).__init__(doc, parent_id, id=id)
        self['columns'] = columns
        el_table = doc.createElement('table', id=self.getRootId())
        doc.getElementById(parent_id).appendChild(el_table)
        el_tr = doc.createElement('tr', id=self.getRootId()+'.tr.header')
        el_table.appendChild(el_tr)
        for col_id, col_name in columns.items():
            el_th = doc.createElement('th', id=self.getRootId()+'.td.header.'+col_id, innerText=col_name)
            el_tr.appendChild(el_th)
    
    def _row_id(self, row_id):
        """
        :returns: the full id of a row
        """
        return self.getRootId()+'.tr.'+row_id

    def _cell_id(self, row_id, col_id):
        return self.getRootId()+'.td.'+row_id+'.'+col_id

    def addRow(self, row_id, values=None):
        "adds a row by keeping rows sorted by row_id"
        _row_id = self._row_id(row_id)
        el_tr = self['doc'].createElement('tr', id=_row_id)
        table_el = self.getElement()
        found = False
        for row_el in table_el.children:
            if row_el.id == self['id']+'.tr.header': continue
            if _row_id < row_el.id:
                found = True
                break
        if found:
            table_el.insertBefore(el_tr, row_el)
        else:
            table_el.appendChild(el_tr)
        i = 0
        for col_id in self['columns']:
            value = "" if values is None else values[i] if type(values) is list else values[col_id] if type(values) is dict else throw(type(values))
            cell_id = self._cell_id(row_id, col_id)
            el_tr.appendChild(self['doc'].createElement('td', id=cell_id, innerText=value))
            i += 1

    def updateRow(self, row_id, values):
        i = 0
        for col_id in self['columns']:
            value = values[i] if type(values) is list else values[col_id] if type(values) is dict else throw(type(values))
            self.updateCell(row_id, col_id, value)
            i+=1

    def addOrUpdateRow(self, row_id, values):
        if self.hasRow(row_id):
            self.updateRow(row_id, values)
        else:
            self.addRow(row_id, values)

    def hasRow(self, row_id):
        return self._row_id(row_id) in self['doc']['elements_by_id']

    def removeRow(self, row_id):
        self['doc'].getElementById(self._row_id(row_id)).remove()

    def getRowIds(self):
        table_el = self['doc'].getElementById(self['id'])
        row_ids = []
        id_prefix = self['id']+'.tr.'
        for row_el in table_el.children:
            if row_el.id != self['id']+'.tr.header':
                row_id = row_el.id
                assert row_id.startswith(id_prefix)
                row_id = row_id[len(id_prefix):]
                row_ids.append(row_id)
        return row_ids

    def updateCell(self, row_id, col_id, value):
        self.getCellElement(row_id, col_id).innerText = value

    def getTableElement(self):
        return self.getElement()

    def getRowElement(self, row_id):
        row_id = self._row_id(row_id)
        return self['doc'].getElementById(row_id)

    def getCellElement(self, row_id, col_id):
        cell_id = self._cell_id(row_id, col_id)
        return self['doc'].getElementById(cell_id)
